# Perfect Song - bass.py
#
# Copyright (C) 2015 Guillaume Tucker <guillaume@mangoz.org>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import splat
import splat.tools.make
import splat.data
import splat.seq
import splat.gen
import splat.scales
import splat.filters
import splat.interpol
import settings
from splat import dB

DEPS = ['settings.py']

class Bar(splat.seq.Pattern):

    def __init__(self, *args, **kw):
        super(Bar, self).__init__(*args, **kw)

        T = 60.0 / settings.TEMPO
        envelope_spline = splat.interpol.spline([
                (0.0, dB(0.0)), (T/4, dB(2.5), 0.0),
                (T/2, dB(-2.0), 0.0), (T*4.1, dB(-0.0))])
        envelope_frag = splat.data.Fragment(rate=settings.RATE, channels=1,
                                            duration=T*2.0)
        envelope_frag.offset(1.0)
        envelope_frag.amp(envelope_spline.signal)
        filters = [(splat.filters.linear_fade, (settings.BASS_DELAY,)),
                   (lambda frag, e: frag.amp(e), (envelope_frag,))]

        self._tri = splat.gen.TriangleGenerator()
        self._tri.filters = filters

        self._gen = splat.gen.OvertonesGenerator()
        self._gen.ot_decexp(1.3, 24)
        self._gen.overtones += \
            ((n * 1.01, 0.0, dB(-18.0 - (n * 1.15))) for n in range(18, 48))
        self._gen.filters = filters

        self._s = splat.scales.LogScale()

    def note(self, frag, t, T, note, lvl=1.0, k=1.0):
        f = self._s[note] * k / 8.0
        self._gen.frag = frag
        self._gen.run(t, t + T, f, levels=lvl)
        self._tri.frag = frag
        lvl_tri = splat.lin2dB(lvl) - dB(4.0)
        self._tri.run(t, t + T, f * 2, levels=dB(lvl_tri))


class Intro(Bar):

    def __init__(self, key, *args, **kw):
        super(Intro, self).__init__(*args, **kw)
        self._key = key
        self._spline = settings.intro_spline()

    def play(self, frag, bar, beat, t, T):
        level = self._spline.value(t)
        if bar == 3 and beat == 3:
            self.note(frag, t, T, self._key, level)
        else:
            self.note(frag, t, T/2, self._key, level)
            self.note(frag, t + T/2, T/2, self._key, level)


class BarA(Bar):

    def __init__(self, notes, *args, **kw):
        super(BarA, self).__init__(*args, **kw)
        self._notes = notes

    def play(self, frag, bar, beat, t, T):
        note = self._notes[beat]
        if note is not None:
            lvl = -1.0 if beat % 2 else 0.0
            self.note(frag, t, T, note, dB(lvl))


class BarB(BarA):

    def play(self, frag, bar, beat, t, T):
        if beat % 2:
            self.note(frag, t, T/2, self._notes[beat], dB(-3), 1.5)
        else:
            self.note(frag, t, T/2, self._notes[beat], dB(0))
        self.note(frag, t + T/2, T/2, self._notes[beat], dB(-3))


class BarC(BarA):

    def play(self, frag, bar, beat, t, T):
        b = beat * 2
        note1 = self._notes[b]
        note2 = self._notes[b + 1]
        if note1 is not None:
            self.note(frag, t, T/2, note1, dB(0))
        if note2 is not None:
            self.note(frag, t + T/2, T/2, note2, dB(-3))


def run(argv):
    frag = splat.data.Fragment(rate=settings.RATE)

    I = Intro('A')
    intro = [I] * settings.INTRO_LENGTH

    V1 = BarA(['E', 'D', 'D', 'D'])
    V2 = BarA(['A', 'A', 'A', 'A'])
    V3 = BarA(['E', 'E', 'D', 'D'])
    V4 = BarA(['A', 'A', 'B', 'B'])
    V5 = BarA(['B', 'B', 'B', 'B'])
    v = lambda x:[V1, V2] * 2 + [V1, BarA(['A', 'A', 'A', x])] + [V3, V4, V5]

    C1 = BarB(['E', 'E', 'D', 'D'])
    C2 = BarA(['B', 'B', 'B', 'B'])
    C3 = [BarC(['E'] * 3 + ['A'] * 5),
          BarC(['A'] * 4 + ['F#-1'] * 4),
          BarC(['B'] * 8),
          BarC(['B'] * 8)]
    chorus = [C1, C2] * 4 + C3

    # Intro
    notes = intro
    # Verse 1
    notes += v(None)
    # Chorus
    notes += chorus
    # Verse 2
    notes += v(None)
    # Chorus
    notes += chorus
    # Bridge
    notes += v('A')
    # Chorus
    notes += chorus
    # Verse 1 again
    notes += v(None)
    # Chorus
    notes += chorus

    splat.seq.PatternSequencer(settings.TEMPO).run(frag, notes)
    return frag

if __name__ == '__main__':
    splat.tools.make.main_frag(sys.argv, run)
