# Perfect Song - qwirks.py
#
# Copyright (C) 2015 Guillaume Tucker <guillaume@mangoz.org>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import math
import random
import splat
import splat.tools.make
import splat.data
import splat.gen
import splat.seq
import splat.scales
import splat.interpol
import settings
from splat import dB

DEPS = ['settings.py']

class Bar(splat.seq.Pattern):

    def __init__(self, *args, **kw):
        super(Bar, self).__init__(*args, **kw)

        T = 60.0 / settings.TEMPO
        sig_len = T * 30
        envelope_spline = splat.interpol.spline([
                (0.0, dB(-24)), (T/50, dB(0), 0.0), (T/10, dB(-6), -0.5),
                (T/3, dB(-9), -0.1), (sig_len+0.1, dB(-96.0))])
        envelope_frag = splat.data.Fragment(rate=settings.RATE, channels=1,
                                            duration=sig_len)
        envelope_frag.offset(1.0)
        envelope_frag.amp(envelope_spline.signal)
        filters = [(splat.filters.linear_fade, (0.001,)),
                   (lambda frag, e: frag.amp(e), (envelope_frag,))]

        def ph(a, k):
            sig_frag = splat.data.Fragment(rate=settings.RATE, channels=1,
                                           duration=sig_len)
            splat.sources.sine(sig_frag, a/1000.0, k)
            return sig_frag

        self._gen = splat.gen.OvertonesGenerator()
        self._gen.overtones = [
            (1.0, ph(0.7, 3.0), dB(0.0)),
            (2.0, ph(0.4, 3.1), dB(-12.0)),
            (4.0, ph(0.2, 6.3), dB(-18.0)),
            (8.0, ph(0.1, 9.4), dB(-40.0)),
            ]
        self._gen.overtones += list(
            ((float(x), 0.0, dB(-24.0 - x*2.1))
             for x in range(3, 48, 3)))
        self._gen.overtones += list(
            ((float(x), 0.0, dB(-18.0 - x*3.0))
             for x in range(5, 35, 5)))
        self._gen.filters = filters

        self._s = splat.scales.LogScale()

    def _note(self, t0, tlen, note, levels):
        self._gen.run(t0, t0+tlen, self._s[note] / 2, levels=levels)


class Intro(Bar):

    def __init__(self, chord, *args, **kw):
        super(Intro, self).__init__(*args, **kw)
        self._chord = chord
        self._spline = settings.intro_spline()

    def play(self, frag, bar, beat, t, T):
        level = splat.lin2dB(self._spline.value(t)) - 2.0
        self._gen.frag = frag
        for n in self._chord:
            if beat == 0:
                self._note(t, T/2, n, dB(level))
            else:
                self._note(t, T/4, n, dB(level))
            self._note(t+T/2, T/4, n, dB(level-3))


class BarA(Bar):

    def __init__(self, chords, *args, **kw):
        super(BarA, self).__init__(*args, **kw)
        self._chords = chords

    def play(self, frag, bar, beat, t, T):
        self._gen.frag = frag
        chord = self._chords[beat]
        if chord is not None:
            notes, nlen = chord
            for n in notes:
                self._note(t, nlen*T, n, dB(0))


class BarB(BarA):

    def play(self, frag, bar, beat, t, T):
        self._gen.frag = frag
        chord = self._chords[beat]
        if chord is not None:
            for n in chord:
                self._note(t, T/2, n, dB(0))
                self._note(t+T/2, T/4, n, dB(-3))


class BarC(BarA):

    def play(self, frag, bar, beat, t, T):
        self._gen.frag = frag
        chords = self._chords[beat]
        for i, c in enumerate(chords):
            t0 = t + (i * T/2)
            if c is not None:
                notes, nlen, lvl = c
                for n in notes:
                    self._note(t0, nlen*T/2, n, lvl)


class Bridge(Bar):

    def __init__(self, chords, *args, **kw):
        super(Bridge, self).__init__(*args, **kw)
        self._chords = chords

    def play(self, frag, bar, beat, t, T):
        self._gen.frag = frag
        for t0 in (t, t+T/2):
            chord = self._chords[beat]
            note = random.choice(chord)
            if note is None:
                continue
            t0 += random.choice([-T/4, 0, 0, 0, 0, 0, T/2])
            nlen = 1.0 + random.uniform(-0.25, 0.5)
            lvl = -3.0 + random.uniform(-1.0, 1.0)
            pan = random.uniform(-3.0, 3.0)
            levels = (dB(lvl+pan), dB(lvl-pan))
            self._note(t0, nlen, note, levels)


def run(argv):
    frag = splat.data.Fragment(rate=settings.RATE)

    cA = ['A', 'C#', 'E']
    cB = ['B', 'D#', 'F#']
    cD = ['D', 'F#', 'A']
    cE = ['E', 'G#', 'B']
    cFs = ['F#', 'A#', 'C#']

    S = splat.seq.Silence()
    I = Intro(cA)
    intro = [I] * 4

    V1 = BarA([(cE, 1.0), (cD, 2.0), None, None])
    V2 = BarA([None, (cA, 2.75), None, None])
    V3 = BarA([None, (cA, 1.25), None, None])
    V4 = [BarA([(cE, 1), (cE, 1), (cD, 1), (cD, 1)]),
          BarA([(cA, 1), (cA, 1), (cB, 1), (cB, 1)]),
          BarA([(cB, 0.5), (cB, 0.5), (cB, 0.5), (cB, 0.5)])]
    verse = [V1, V2] * 2 + [V1, V3] + V4

    C1 = BarB([cE, None, cD, None])
    C2 = BarB([cB, cB, cB, cB])
    C3 = [BarC([((['E', 'B', 'E1'], 3.0, dB(-4)), None),
                (None, (['A', 'E', 'A1'], 9.0, dB(-2))),
                (None, None),
                (None, None)]),
          BarC([(None, None),
                (None, None),
                ((['F#-1', 'C#', 'F#'], 4.0, dB(-1)), None),
                (None, None)]),
          BarC([((['B', 'F#', 'B1'], 16.0, dB(0)), None),
                (None, None),
                (None, None),
                (None, None)]),
          S]
    chorus = [C1, C2] * 4 + C3

    BC1 = BarA([(cE, 1.0), (cD, 3.0), None, None])
    BC2 = BarA([(cA, 4.0), None, None, None])
    BC3 = [BarA([(cE, 2.0), None, (cD, 2.0), None]),
           BarA([(cA, 2.0), None, (cB, 2.0), None]),
           BarA([(cB, 4.0), None, None, None])]
    bridge_chords = [BC1, BC2] * 3 + BC3

    bcE1 = ['E1', 'E1', 'G#1', 'B1', 'B1', 'E2', 'E2']
    bcE2 = ['E1', 'E1', 'G#1', 'B1', 'B1', 'E2', 'G#2', None]
    bcE3 = ['E1', 'E1', 'B1', 'B1', 'E2']
    bcD1 = ['D1', 'D1', 'F#1', 'A1', 'A1', 'D2', 'F#2', None]
    bcD2 = ['D1', 'D1', 'A1', 'A1', 'D2']
    bcA1 = ['A1', 'A1', 'C#1', 'E1', 'E1', 'A2', 'C#2', None]
    bcA2 = ['A1', 'A1', 'E1', 'E1', 'A2']
    bcA3 = ['A1', 'A1', 'A2']
    bcB1 = ['B1', 'B1', 'D#1', 'F#1', 'F#1', 'A1', 'B2', 'F#2', None]
    bcB2 = ['B1', 'B1', 'F#1', 'A1', 'B2', None, None]
    bridge = [Bridge([bcE1, bcD1, bcD1, bcD2]),
              Bridge([bcA1, bcA1, bcA1, bcA2]),
              Bridge([bcE2, bcD1, bcD1, bcD2]),
              Bridge([bcA1, bcA1, bcA1, bcA2]),
              Bridge([bcE2, bcD1, bcD1, bcD2]),
              Bridge([bcA1, bcA1, bcA1, bcA2]),
              Bridge([bcE1, bcE3, bcD1, bcD2]),
              Bridge([bcA3, bcA3, bcB1, bcB1]),
              Bridge([bcB1, bcB1, bcB2, bcB2])]

    # Intro
    seq = intro
    # Verse 1
    seq += verse
    # Chorus
    seq += chorus
    # Verse 2
    seq += verse
    # Chorus
    seq += chorus
    # Bridge
    seq += zip(bridge_chords, bridge)
    # Chorus
    seq += chorus
    # Verse 1 again
    seq += verse
    # Chorus
    seq += chorus

    splat.seq.PatternSequencer(settings.TEMPO).run(frag, seq)
    return frag

if __name__ == '__main__':
    splat.tools.make.main_frag(sys.argv, run)
